/* $Id: XtXcgrp.cpp,v 1.1 2006/02/02 15:37:57 jsato Exp $ */
/*
 * Copyright (C) 2006 WIDE Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * 2006/02/03 jsato: created
 *
 */


#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <arpa/inet.h>

#include <qtimer.h>
#include <qdns.h>
#include <qhttp.h>
#include <qprocess.h>
#include <qregexp.h>

#include "qxcgroup.h"
#include "XtHttp.h"
#include "XtXcgrp.h"



XtXcgrp::XtXcgrp(XtConf *conf)
{
  xtConf = conf;
  grpName    = xtConf->getGroupName();
  httpServer = xtConf->getGroupServer();
  httpPort   = xtConf->getGroupPort();
  ownAddr    = QString("");
  ownPrefix  = 0;
  mbrNum     = 0;

  baseUrl.sprintf(XCGRP_URL, grpName.data());
  mbusAddr.sprintf(XCGRP_MBUS,getpid());

  grpOps = new XtHttp(httpServer, httpPort);

  /* Register SLOT (XtXcgrp -> XtHttp) */
  connect(grpOps, SIGNAL(sigSendEnd()),
          this, SLOT(httpRecvEnd()));
  connect(grpOps, SIGNAL(sigSendErr(Q_UINT8)),
          this, SLOT(httpRecvErr(Q_UINT8)));

  /* Register SLOT (XtHttp -> XtXcgrp) */
  connect(this, SIGNAL(sigUpdateHttp(QString, Q_UINT16)),
          grpOps, SLOT(httpUpdate(QString, Q_UINT16)));

  grpGet = new XtHttp(httpServer, httpPort);

  /* Register SLOT (XtXcgrp -> XtHttp) */
  connect(grpGet, SIGNAL(sigSendData(const char *, Q_ULONG)),
          this, SLOT(httpRecvData(const char *, Q_ULONG)));
  connect(grpGet, SIGNAL(sigSendErr(Q_UINT8)),
          this, SLOT(httpRecvErr(Q_UINT8)));

  /* Register SLOT (XtHttp -> XtXcgrp) */
  connect(this, SIGNAL(sigUpdateHttp(QString, Q_UINT16)),
          grpGet, SLOT(httpUpdate(QString, Q_UINT16)));

  mbusHdl = new XtMbus(mbusAddr);
  mbusHdl->keepalive();
}


XtXcgrp::~XtXcgrp()
{
  delete mbusHdl;
  delete grpOps;
  delete grpGet;
}


void XtXcgrp::join()
{
  QString httpUrl = baseUrl;
  httpUrl.append("join&addr=");
  httpUrl.append(ownAddr);

  MSGVRB("%s Join Group %s", get_time().data(), grpName.data());

  grpOps->get(httpUrl, FALSE);
  mbusHdl->keepalive();
  return;
}


void XtXcgrp::leave()
{
  QString httpUrl = baseUrl;
  httpUrl.append("leave&addr=");
  httpUrl.append(ownAddr);

  MSGVRB("%s Leave Group %s", get_time().data(), grpName.data());

  grpOps->stop();
  grpOps->get(httpUrl, FALSE);
  return;
}


void XtXcgrp::show()
{
  QString httpUrl = baseUrl;
  httpUrl.append("show");

  MSGVRB("%s Show Group %s", get_time().data(), grpName.data());

  grpGet->get(httpUrl, TRUE);
  return;
}


void XtXcgrp::httpRecvData(const char *data, Q_ULONG data_len)
{
  grpGet->stop();

  /*
   * compare list of got members with that of current members
   * - add if got member NOT exists in member list
   * - set TRUE to "alive" flag if got member exists in current member list
   * - remain FALSE if there is no member in list of got members though it
   *   exists in current members
   */
  QString members(data);
  QString preMember("");
  QStringList list = QStringList::split("\n", members, FALSE);
  for(QStringList::Iterator it = list.begin(); it != list.end(); ++it){
    MSGVRB("%s", (*it).data());
    if(is_member(*it) == FALSE){
      if(preMember.isEmpty() == TRUE)
        add_member(*it);
      else
        addto_member(*it, preMember);
    }else{
      XtMember *mbr = get_member((*it));
      mbr->alive = TRUE;
    }
    preMember = (*it);
  }

  /*
   * update member list based on "alive" flag
   * - if TRUE,  remain in member list and reset FALSE
   * - if FALSE, remove from member list
   *  and, make mbus members based on "send" flag
   */
  members = "";
  QValueList<XtMember>::iterator it = mbrList.begin();
  while(it != mbrList.end()){
    MSGDBG("%s, send=%s, alive=%s, scope=%d",
           (*it).addr.data(),
           (*it).send  == TRUE ? "yes" : "no",
           (*it).alive == TRUE ? "yes" : "no", (*it).scope);

    if((*it).alive == TRUE){
      (*it).alive = FALSE;
      if((*it).send == TRUE){
        QString tmp("");
        tmp.sprintf(",%s", (*it).addr.data());
        members += tmp;
      }
      ++it;
    }else{
      QValueList<XtMember>::iterator del = it;
      ++it;
      emit sigDelMember((*del).addr);
      mbrList.erase(del);
    }
  }

  Q_UINT8 num = mbrList.count();
  if(mbrNum < num)
    mbrNum = num;
  else if(mbrNum > num){
    mbrNum = num;
#ifdef QXCGROUP_MEMBER_SORT
    sort_member();
    emit sigUpdateMember(mbrList);
#endif /* QXCGROUP_MEMBER_SORT */
  }

  if(members.isEmpty() == FALSE){
    members.remove(0, 1);
    QString mbusMember("");
    mbusMember.sprintf("\"%s\"", members.data());
    mbusHdl->announce(mbusMember);

    MSGDBG("Mbus Member: %s", mbusMember.data());
  }

  return;
}


void XtXcgrp::httpRecvEnd()
{
  grpOps->stop();
  return;
}


void XtXcgrp::add_member(QString member)
{
  if(is_member(member) == TRUE){
    /* Member already exists */
    emit sigSendErr(3);
    return;
  }

  XtMember new_mbr;
  new_mbr.addr  = member;
  new_mbr.send  = TRUE;
  new_mbr.alive = TRUE;
  new_mbr.scope = get_scope(member);
  mbrList.push_back(new_mbr);

#ifdef QXCGROUP_MEMBER_SORT
  sort_member();
  emit sigUpdateMember(mbrList);
#else
  emit sigAddMember(new_mbr);
#endif /* QXCGROUP_MEMBER_SORT */
 
  return;
}


void XtXcgrp::addto_member(QString member, QString pre)
{
  if(is_member(member) == TRUE){
    /* Member already exists */
    emit sigSendErr(3);
    return;
  }

  XtMember *mbr = get_member(pre);
  if(mbr == 0){
    /* No such member found */
    emit sigSendErr(2);
    return;
  }

  XtMember new_mbr;
  new_mbr.addr  = member;
  new_mbr.send  = TRUE;
  new_mbr.alive = TRUE;
  new_mbr.scope = get_scope(member);

  /* JSATO - the following does not work :-p
   *         currently, add new member to the end of member list even though 
   *         it NOT exists in the end of got list
   */
#if 0
  if(mbrList.last() == *mbr){
    MSGDBG("add next to: %s", mbr->addr.data());
    mbrList.push_back(new_mbr);
  }else{
    QValueList<XtMember>::iterator it;
    it++;
    MSGDBG("add in front of : %s", (*it).addr.data());
    mbrList.insert(it, new_mbr);
  }
#endif

  mbrList.push_back(new_mbr);

#ifdef QXCGROUP_MEMBER_SORT
  sort_member();
  emit sigUpdateMember(mbrList);
#else
  emit sigAddToMember(new_mbr, *mbr);
#endif /* QXCGROUP_MEMBER_SORT */
 
  return;
}


void XtXcgrp::del_member(QString member)
{
#if 0
  XtMember *mbr = get_member(member);
  if(mbr == 0){
    //no such member found
    emit sigSendErr(2);
    return;
  }

  mbrList.remove(*mbr);
  emit sigDelMember(member);
#endif
  return;
}


void XtXcgrp::un_member(QString member, bool flag)
{
  XtMember *mbr = get_member(member);
  if(mbr == 0){
    /* No such member found */
    emit sigSendErr(2);
    return;
  }

  mbr->send = flag;

  return;
}


bool XtXcgrp::is_member(QString member)
{
  bool found = FALSE;
  QValueList<XtMember>::iterator it;
  for(it = mbrList.begin(); it != mbrList.end(); ++it){
    if((*it).addr == member)
      found = TRUE;
  }
  return found;
}


XtMember* XtXcgrp::get_member(QString member)
{
  bool found = FALSE;
  QValueList<XtMember>::iterator it;
  for(it = mbrList.begin(); it != mbrList.end(); ++it){
    if((*it).addr == member){
      found = TRUE;
      break;
    }
  }

  if(found == FALSE)
    return 0;

  return (XtMember *)&(*it);
}



void XtXcgrp::confMember(XtMember member)
{
  XtMember *mbr = get_member(member.addr);
  if(mbr == 0){
    /* no such member found */
    emit sigSendErr(2);
    return;
  }

  mbr->send = member.send;

  return;
}


void XtXcgrp::confUpdate()
{
  grpName    = xtConf->getGroupName();
  httpServer = xtConf->getGroupServer();
  httpPort   = xtConf->getGroupPort();
  ownAddr    = xtConf->getIfAddr();
  ownPrefix  = xtConf->getIfPrefix();

  if(ownAddr.isEmpty() == FALSE){
    memset(&ownSock, 0, sizeof(struct sockaddr_in6));
    inet_pton(AF_INET6, (const char *)ownAddr.data(), &ownSock.sin6_addr);
    ownSock.sin6_scope_id = ownPrefix;
  }

  emit sigUpdateHttp(httpServer, httpPort);

  return;
}


QString XtXcgrp::get_time()
{
  QTime now   = QTime::currentTime();
  QDate today = QDate::currentDate();
  QString time(today.toString("ddd MMM d"));
  time.append(" ");
  time.append(now.toString("hh:mm:ss"));
  time.append(" ");
  time.append(today.toString("yyyy"));

  return time;
}


Q_UINT8 XtXcgrp::get_scope(QString addr)
{
  int i, j, base = 0, offset = 0, matched = 0;
  struct sockaddr_in6 sock6;

  if(ownAddr == addr){
    MSGDBG("%s is Own", addr.data());
    return (Q_UINT8)MBR_SCOPE_OWN;
  }

  inet_pton(AF_INET6, (const char *)addr.data(), &sock6.sin6_addr);
  base   = (ownSock.sin6_scope_id-1) / 8;
  offset = (ownSock.sin6_scope_id-1) % 8;

  for(i = 0; i <= base; i++){
    if(i == base){
      for(j = 0; j <= offset; j++){
        if((sock6.sin6_addr.s6_addr[i] & ~(0xff >> (j+1))) !=
           (ownSock.sin6_addr.s6_addr[i] & ~(0xff >> (j+1))))
          goto out;
        else
          matched++;
      }
out:
      /* JSATO - compare number of matched with total bits(e.g. offset+1)
       *         offset start with "0", so it plus 1 as total bits
       */
      if(matched == (offset+1)){    /* all bit matched */
        MSGDBG("%s is Same Link", addr.data());
        return (Q_UINT8)MBR_SCOPE_LINK;
#ifdef QXCGROUP_MEMBER_SORT
      }else if(matched > 4){        /* at least 4 bit matched */
        MSGDBG("%s is Same Site", addr.data());
        return (Q_UINT8)MBR_SCOPE_SITE;
#endif /* QXCGROUP_MEMBER_SORT */
      }
    }

    if((sock6.sin6_addr.s6_addr[i] & 0xff) != ownSock.sin6_addr.s6_addr[i])
      break;
  }

  MSGDBG("%s is Other Network", addr.data());
  return (Q_UINT8)MBR_SCOPE_OTHER;
}


void XtXcgrp::httpRecvErr(Q_UINT8 ret)
{
  switch(ret){
    case XCGRP_OK:
      break;
    case XCGRP_ERR:
      MSGERR("Error has occured");
      break;
    case XCGRP_E_NO_MBR:
      MSGERR("No Such Member");
      break;
    case XCGRP_E_MBR_EXIST:
      MSGERR("Member Already Exists");
      break;
    case XCGRP_E_SOCKET:
      MSGERR("Unable to create socket");
      break;
    case XCGRP_E_CONNECT:
      MSGERR("Unable to connect");
      break;
    case XCGRP_E_SEND:
      MSGERR("Send Error");
      break;
    case XCGRP_E_RECV:
      MSGERR("Receive Error");
      break;
    case XCGRP_E_BAD_ADDR:
      MSGERR("Invalid Address format");
      break;
    case XCGRP_E_NO_SOCK:
      MSGERR("No socket created");
      break;
    case XCGRP_E_HTTP:
      MSGERR("HTTP Access Error");
      break;
    case XCGRP_E_HTTP_400:
      MSGERR("HTTP 400 Bad Request");
      break;
    case XCGRP_E_HTTP_401:
      MSGERR("HTTP 401 Unauthorized");
      break;
    case XCGRP_E_HTTP_402:
      MSGERR("HTTP 402 Payment Required");
      break;
    case XCGRP_E_HTTP_403:
      MSGERR("HTTP 403 Forbidden");
      break;
    case XCGRP_E_HTTP_404:
      MSGERR("HTTP 404 Not Found");
      break;
    case XCGRP_E_HTTP_405:
      MSGERR("HTTP 405 Method Not Allowed");
      break;
    case XCGRP_E_HTTP_406:
      MSGERR("HTTP 406 Not Acceptable");
      break;
    case XCGRP_E_HTTP_407:
      MSGERR("HTTP 407 Proxy Authentication Required");
      break;
    case XCGRP_E_HTTP_408:
      MSGERR("HTTP 408 Request Timeout");
      break;
    case XCGRP_E_HTTP_500:
      MSGERR("HTTP 500 Internal Sever Error");
      break;
    case XCGRP_E_HTTP_501:
      MSGERR("HTTP 501 Not Implemented");
      break;
    case XCGRP_E_HTTP_502:
      MSGERR("HTTP 502 Bad Gateway");
      break;
    case XCGRP_E_HTTP_503:
      MSGERR("HTTP 503 Service Unavailable");
      break;
    case XCGRP_E_HTTP_504:
      MSGERR("HTTP 504 Gateway Timeout");
      break;
    case XCGRP_E_HTTP_505:
      MSGERR("HTTP 505 HTTP/Version Not Supported");
      break;
    case XCGRP_E_HTTP_BUSY:
      MSGERR("HTTP Server Too Busy");
      break;
    default:
      MSGERR("Unknown Error: %d", ret);
      break;
  }
  return;
}


#ifdef QXCGROUP_MEMBER_SORT
void XtXcgrp::sort_member()
{
  MSGDBG("XtXcgrp::sort_member() called");

  /*
   * Resort member list based on "scope" value
   *  sorting order
   *   1. own address
   *   2. same link
   *   3. same site 
   *   4. other network 
   */
  QValueList<XtMember> new_list;
  QValueList<XtMember>::iterator it;

  /* Get own address */
  for(it = mbrList.begin(); it != mbrList.end(); ++it)
    if((*it).scope == MBR_SCOPE_OWN)
      new_list.push_back((*it));

  /* Get member of same link */
  for(it = mbrList.begin(); it != mbrList.end(); ++it)
    if((*it).scope == MBR_SCOPE_LINK)
      new_list.push_back((*it));

  /* Get member of same site */
  for(it = mbrList.begin(); it != mbrList.end(); ++it)
    if((*it).scope == MBR_SCOPE_SITE)
      new_list.push_back((*it));

  /* Get member of other network */
  for(it = mbrList.begin(); it != mbrList.end(); ++it)
    if((*it).scope == MBR_SCOPE_OTHER)
      new_list.push_back((*it));

  if(new_list.count() != mbrList.count()){
    MSGVRB("Member resorting error");
    return;
  }

  mbrList = new_list;

  return;
}
#endif /* QXCGROUP_MEMBER_SORT */
